import sys
import math

#
# Control constants
#

OCTAVES = 7                # number of octaves supported
MIDDLE_C = 3 * 12          # note number of middle C
TUNING_NOTE = MIDDLE_C + 9 # A above middle C
TUNING_FREQ = 440.0        # A = 440hz
GLISS_MAX = 8              # largest allowed gliss = GLISS_MAX-1
SWING = 1                  # how many frames to swing sax 16ths by
GRACE_LEN = 2              # length of grace note
BASS_SHORTEN = 3           # length to shorten every bass note
ARP_LEN = 2                # length of arpeggio note in chords
BAR_LEN = 24 * 4           # frames in a bar
SECTION_BARS = 8           # bars in a section
SECTIONS = 21              # sections

#
# File output
#

f_out = None
f_log = None

def open_files(outfile, logfile):
    global f_out
    global f_log
    f_out = open(outfile, "w")
    f_log = open(logfile, "w")

def close_files():
    global f_out
    global f_log
    f_out.close()
    f_log.close()
    
def log(s):
    f_log.write(s)
    sys.stdout.write(s)
    
def header(s):
    f_out.write(s)
    #sys.stdout.write(s)

#
# Lilypond reading
#

class Note:
    """Members:
        * pitch (36 = C, 0 = rest, 1 tie)
        * duration (6 = 16th, 0 = grace, 24 = 4tr)
        * gliss (starting pitch, 0 if no gliss)
        * ghost (True if ghosted)
        * chord (pitch list if chord, pitch member will be highest pitch)
    """
    def __init__(self, pitch, duration, gliss, ghost):
        self.pitch = pitch
        self.duration = duration
        self.gliss = gliss
        self.ghost = False
        self.chord = None

class LyReader:
    """Parses a .ly file.
        Members:
        * tokens - list of whitespace separated tokens from the file (internal)
        * position - index into tokens (internal)
        * duration - current duration
        * times_n/times_d - note duration = duration
        * tied - whether or not to expect a tie
        * cadenza - cadenza mode on
        * debug - whether to debug all advanced tokens
    """
    def __init__(self, filename):
        self.debug = False
        self.tokens = []
        lines = open(filename, "rt").readlines()
        for line in lines:
            for token in line.split():
                if token[0] == "%": # do not read comments
                    break;
                self.tokens.append(token)
        self.reset()

    def reset(self):
        """Resets position to beginning of file,
            resets other internal states."""
        self.position = 0
        self.duration = 0
        self.times_n = 1
        self.times_d = 1
        self.tied = False
        self.cadenza = False

    def advance(self):
        """Reads the next token and advances position."""
        if (self.position >= len(self.tokens)):
            return None
        token = self.tokens[self.position];
        if self.debug:
            log("advance() = \"%s\"\n" % token)
        self.position += 1
        return token

    def seek(self, token):
        """Seeks to next instance of token and skips it.
            raises exception if failed.
        """
        t = self.advance()
        while (t != token):
            if (t == None):
                raise Exception("seek(\"%s\"): failed to find token." % token);
            t = self.advance()

    def unbrace(self):
        """Seek, skipping matched { } pairs until } is found,
            raises exception if failed.
        """
        braces = 1
        while braces:
            t = self.advance()
            if (t == None):
                raise Exception("unbrace(): Closing brace not found.")
            if (t != "}"):
                if (t == "{"):
                    braces += 1
            else:
                braces -= 1

    def skip(self, token):
        """Like advance() but verifies the token via Exception."""
        t = self.advance()
        if (t != token):
            raise Exception("Skipped token not matched: %s" % t)
        return t

    def read_note_token(self,t):
        """Parses a token into a Note."""
        if self.debug:
            log("Note token: %s\n" % t)
        pitch = 0
        duration = self.duration
        chord = None
        dpos = 0 # where to read duration
        if t[0] == 'r':
            pitch = 0
            dpos = 1            
        elif t[0] >= 'a' or t[0] <= 'g':
            PITCH_CLASS = {'c':0, 'd':2, 'e':4, 'f':5, 'g':7, 'a':9, 'b':11}
            pitch = PITCH_CLASS[t[0]]
            pitch += MIDDLE_C - 12
            dpos = 1
            # flats
            while len(t) > dpos and t[dpos] == 'f':
                pitch -= 1
                dpos += 1
            # sharps
            while len(t) > dpos and t[dpos] == 's':
                pitch += 1
                dpos += 1
            # octave
            while len(t) > dpos and t[dpos] == ',':
                pitch -= 12
                dpos += 1
            while len(t) > dpos and t[dpos] == '\'':
                pitch += 12
                dpos += 1
        else:
            raise Exception("read_note(): Unable to parse pitch token: %s" % t)
        # skip force-accidental notation
        if len(t) > dpos and t[dpos] == '!':
            dpos += 1
        # skip ~ if attacked to a note (I want to repeat these notes anyway)       
        if len(t) > dpos and t[dpos] == '~':
            dpos += 1
        # skip end of chord > if in a token
        if len(t) > dpos and t[dpos] == '>':
            chord = [0] # signals chord, resolved in read_note()
            dpos += 1
        # read duration if attached
        if len(t) > dpos and t[dpos] >= '0' and t[dpos] <= '9':
            division = int(t[dpos]) - int('0')
            dpos += 1
            while len(t) > dpos and t[dpos] >= '0' and t[dpos] <= '9':
                division *= 10
                division += int(t[dpos]) - int('0')
                dpos += 1
            DIVISION = { 1:96, 2:48, 4:24, 8:12, 16:6, 32:3 }
            duration = DIVISION[division]
            dot = duration
            while len(t) > dpos and t[dpos] == '.':
                dot = dot / 2
                duration += dot
                dpos += 1
        # apply ties (hope they match!)
        if (self.tied):
            pitch = 1
        self.tied = False
        # apply duration
        self.duration = duration
        duration = (duration * self.times_n) / self.times_d
        # return finished note
        note = Note(pitch, duration, 0, False)
        note.chord = chord
        return note

    def read_note(self):
        """Reads the next note from a block.
            User should advance past { at the block start before beginning.
            When } is reached read_note will return None
        """
        t = self.advance()
        # ignore some tokens (beaming [], barline |)
        if     t == "[" \
            or t == "]" \
            or t == "|" :
            return self.read_note()
        # do not allow some tokens
        if t == "{":
            raise Exception("Unable to parse token: %s" % t)
        # close brackets mark end of notes, or
        # end of \times
        if t == "}":
            if self.times_n != 1 or self.times_d != 1:
                self.times_n = 1
                self.times_d = 1
                return self.read_note()
            elif self.cadenza:
                self.skip("\\cadenzaOff")
                self.cadenza = False
                return self.read_note()
            else:
                return None # end of block is reached
        # skip markup
        if t == "^":
            self.skip("\\markup")
            self.skip("{")
            self.unbrace()
            return self.read_note()
        # tied notes
        if t == "~":
            self.tied = True
            return self.read_note()
        # ghosted notes
        if t == "\\ghost":
            note = self.read_note()
            note.ghost = True
            return note
        # gliss notes
        if t == "\\glissin":
            note = self.read_note()
            gliss = note.pitch
            note = self.read_note()
            note.gliss = gliss
            return note
        # grace notes
        if t == "\\grace":
            note = self.read_note()
            note.duration = 0
            return note
        # times
        if t == "\\times":
            t = self.advance()
            self.times_n = int(t[0]) - int('0')
            self.times_d = int(t[2]) - int('0')
            if (t[1] != '/'):
                raise Exception("read_note(): improperly formatted \\times expression.")
            self.skip("{")
            return self.read_note()
        # cadenza
        if t == "\\cadenzaOn":
            self.skip("\\grace")
            self.skip("{")
            self.cadenza = True
            return self.read_note()
        # chords
        if t[0] == '<':
            chord = []
            chord_note = self.read_note_token(t[1:])
            chord_tied = (chord_note.pitch == 1)
            chord_ghost = chord_note.ghost
            note = Note(0,0,0,False)
            while True:
                chord.append(chord_note.pitch)
                note.duration = chord_note.duration
                note.pitch = chord_note.pitch
                if (chord_note.chord != None): # chord end signalled
                    break
                chord_note = self.read_note()
            if chord_tied:
                note.chord = None
                note.pitch = 1
            else:
                note.chord = chord
            note.ghost = chord_ghost
            return note
        # anything that has made it here is presumed a note
        note = self.read_note_token(t)
        if self.cadenza:
            note.duration = 4
        return note

#
# Music generation
#

pitch_symbol = {}
REST_SYMBOL  = "RES"
TIE_SYMBOL   = "TIE"
GHOST_SYMBOL = "GHO"
GLISS_SYMBOL = "SLI"
BAR_SYMBOL   = "BAR"
REST  = 0
TIE   = 1
GHOST = 2
GLISS = 3
BAR   = 4

def define_symbols():
    global pitch_symbol
    # build symbol table
    pitch_symbol = dict()
    pitch_symbol[REST ] = REST_SYMBOL
    pitch_symbol[TIE  ] = TIE_SYMBOL
    pitch_symbol[GHOST] = GHOST_SYMBOL
    pitch_symbol[GLISS] = GLISS_SYMBOL
    pitch_symbol[BAR  ] = BAR_SYMBOL
    for i in range(12,12*(OCTAVES+1)):
        octave = math.floor((i - 12) / 12)
        pitch_class = (i % 12)
        PITCH_CLASS = ["C_","Cs","D_","Ds","E_","F_",\
                       "Fs","G_","Gs","A_","As","B_"]
        name = "%s%d" % (PITCH_CLASS[pitch_class], octave)
        pitch_symbol[i] = name
    # output symbol table
    header("// defines\n")
    header("#define LOOP 127\n")
    for (i,n) in pitch_symbol.items():
        if i == 12:
            header("\n// temporary note data names\n")
        header("#define %s  %3d\n" % (n,i))
    header("\n")

def undefine_symbols():
    header ("// undefine temporary note data names\n")
    for (i,n) in pitch_symbol.items():
        if (i >= 12):
            header("#undef %s\n" % (n))
    header("\n")

def swing_16ths(notes):
    total_dur = 0
    for i in range(0,len(notes)):
        dur = notes[i].duration
        if ((total_dur % 12) == 6) and (notes[i].duration >= 6):
            notes[i].duration -= SWING # steal time from note
            j = i-1
            while notes[j].duration < 1: # skip grace notes
                j -= 1
            notes[j].duration += SWING # give time to previous note
        total_dur += dur

def shorten_bass(notes):
    """
    Shortens every bass note by BASS_SHORTEN.
    Note that ties between sections are not accounted for.
    """
    new_notes = []
    for i in range(0,len(notes)):
        n = notes[i]
        o = Note(REST,0,0,False) # fake rest for end of segment
        if ((i+1) < len(notes)):
            o = notes[i+1]
        if n.pitch == TIE or n.pitch >= 12:
            if o.pitch >= 12 or o.pitch == REST:
                n.duration -= BASS_SHORTEN
                new_notes.append(n)
                new_notes.append(Note(REST,BASS_SHORTEN,0,False))
            else:
                new_notes.append(n)
        else:
            new_notes.append(n)
    return new_notes

def expand_grace(notes):
    for i in range(0,len(notes)):
        if (notes[i].duration == 0):
            notes[i-1].duration -= GRACE_LEN
            notes[i].duration = GRACE_LEN

def build_part(ly, part, start, stop, suppress_table, chords):
    header("// %s\n\n" % (part))
    ly.reset()
    for s in range(start,stop):
        section = "%s%c" % (part, chr(s + ord('A')))
        # read data
        ly.seek(section)
        ly.skip("=")
        ly.skip("{")
        notes = []
        total_dur = 0
        n = ly.read_note()
        while (n != None):
            notes.append(n)
            total_dur += n.duration
            n = ly.read_note()
        log("%s read: %d\n" % (section, total_dur))
        # format data
        if (part == "sax"):
            swing_16ths(notes)
        expand_grace(notes)
        if (part == "bass"):
            notes = shorten_bass(notes)
            if (s == SECTIONS-1): # cancel shorten on last note
                notes = notes[0:len(notes)-1]
                notes[len(notes)-1].duration += BASS_SHORTEN
        # output data
        pos = 0
        bars = []
        total_dur = 0
        for b in range(0,SECTION_BARS):
            bar = "%s%d" % (section, b)
            bars.append(bar)
            header("const uint8 %s[] = {" % (bar))
            dur = 0
            # the last sax bar gets an exceptional length for its cadenza
            cadenza = ((part=="sax") and (b==(SECTION_BARS-1)) and (s==(SECTIONS-1)))
            while (pos < len(notes)) and (dur < BAR_LEN or cadenza):
                n = notes[pos]
                pos += 1
                dur += n.duration
                if (n.ghost == True):
                    header("%s,    " % pitch_symbol[GHOST])
                if (n.gliss != 0):
                    header("%s,%s," % (pitch_symbol[GLISS], pitch_symbol[n.gliss]))
                    gliss_width = n.pitch - n.gliss
                    if gliss_width >= GLISS_MAX:
                        raise Exception("Gliss too wide (%d). Increase GLISS_MAX." % (gliss_width))
                if (n.chord == None) or (n.pitch == TIE):
                    # pitch, duration
                    header("%s,%3d," % (pitch_symbol[n.pitch], n.duration))
                else:
                    # chord #, duration
                    header("%3d,%3d," % (len(chords), n.duration))
                    # store the chord
                    chords.append(n.chord)
            header("%s};\n" % (pitch_symbol[BAR]))
            log("%s output: %d\n" % (bar, dur))
            if (dur != BAR_LEN and not cadenza):
                raise Exception("Part %s bar %d/%d is too long." % (part,s,b))
            total_dur += dur
        header("const uint8* const %s[] = { " % (section))
        for bar in bars:
            header("%s, " % (bar))
        header("};\n\n")
        log("%s output: %d\n" % (section, total_dur))
    if not suppress_table:
        header("const uint8* const * const %s[] = { " % (part))
        for s in range(0,SECTIONS):
            if s >= start and s < stop:
                section = "%s%c" % (part, chr(s + ord('A')))
            else:
                section = "NULL"
            header("%s, " % (section))
        header("};\n\n")

def build_sax(ly):
    build_part(ly,"sax",0,SECTIONS,False,None)

def build_bass(ly):
    build_part(ly,"bass",0,SECTIONS,False,None)

def build_pianoT(ly):
    build_part(ly,"pianoT",13,17,False,None)

def build_pianoB(ly,chords):
    build_part(ly,"pianoB", 0, 1,True,chords)
    build_part(ly,"pianoB",13,17,False,chords)

def build_chords(chords):
    header("// chord macros\n")
    for i in range(0,len(chords)):
        header("const sint8 chord%d[] = {" % (i))
        chord = chords[i]
        if (len(chord) == 0):
            header("0,")
        for n in chord:
            for j in range(0,ARP_LEN):
                header("%s," % (pitch_symbol[n]))
        header("LOOP,%d};\n" % (ARP_LEN*len(chord)))
    header("const sint8* const chord[] = { ")
    for i in range(0,len(chords)):
        header("chord%d, " % (i))
    header("};\n\n")
    log("Chord macros generated: %d\n" % (len(chords)))
    
def build_gliss(max_slide):
    header("// sax slide macros\n")
    for i in range(0,max_slide):
        header("const sint8 slide%d[] = {" % (i))
        for j in range(0,i):
            m = j-i           
            header("%d," % (m))
            if (m >= -3): # stretch out end of gliss
                header("%d," % (m))
        header(" 0,LOOP,1};\n")
    header("const sint8* const slide[] = { ")
    for i in range(0,max_slide):
        header("slide%d, " % (i))
    header("};\n\n")
    log("Gliss macros generated: %d\n" % (max_slide))

def build_pitch_table():
    pitches = 12 * (OCTAVES+1)
    header("// pitch table\n")
    header("const uint16 pitch[%d] = {" % (pitches))
    for i in range(0,pitches):
        if ((i%12) == 0):
            header("\n   ")
        # find the desired pitch in Hz
        relative = i - TUNING_NOTE
        freq = TUNING_FREQ * pow(2, relative/12)
        # convert to NES table value
        NESFREQ = 315000000.0 / 176.0
        table = int(round((NESFREQ/(16.0*freq))-1.0))
        header(" 0x%04X," % (table))
    header("\n};\n\n")
    log("Pitch table generated: %d pitches\n" % (pitches))

#
# Main
#

if __name__ == "__main__":
    ly = LyReader("coltrane.ly")
    open_files("music.h", "buildmusic.log")
    #ly.debug = True
    header("// auto generated from coltrane.ly\n")
    header("\n")  
    define_symbols()
    build_sax(ly)
    build_bass(ly)
    build_pianoT(ly) # solo
    header("// model chord progression\n")
    chords = [[],[],[],[],[],[],[],[],[],[],[],[]] # 12 dummy chords
    build_pianoB(ly,chords) # chords
    build_chords(chords)
    build_gliss(GLISS_MAX)
    build_pitch_table()
    undefine_symbols()
    header("// end of file")
    close_files()
